module Build (build) where

import Relude
import Text.Pandoc.Definition
import Text.Pandoc.Class
import Text.Pandoc.Options (def)
import qualified Text.Pandoc.Filter as Filter
import Nebelhorn
import Util
import Config
import Save
import Path
import Data.List (union)
import qualified Data.Map as Map (singleton, empty, insert)
import Control.Concurrent.Async (mapConcurrently)
import qualified Data.Text as T
import Data.Char (isAlphaNum)


mapPages :: (Page -> Page) -> Nebelhorn -> Nebelhorn
mapPages f nebelhorn@Nebelhorn{..} = nebelhorn { nebelhornPages = f <$> nebelhornPages }

applyDoc :: (Pandoc -> Pandoc) -> (Page -> Page)
applyDoc f page@Page{..} = page { pageDoc = f pageDoc }

-- | Puts a path into the meta values of a pandocument.
linkIntoMeta :: Page -> Page
linkIntoMeta page@Page{..} = page
    { pageDoc = addMeta "link" link pageDoc
    } where link = MetaString $ toText $ toFilePath $ [absdir|/|] </> pageTarget

baseUrlIntoMeta :: Nebelhorn -> Nebelhorn
baseUrlIntoMeta nebelhorn@Nebelhorn{..} = mapPages baseUrlIntoMeta' nebelhorn where

    baseUrlIntoMeta' :: Page -> Page
    baseUrlIntoMeta' page@Page{..} = page
        { pageDoc = addMeta "baseUrl" (MetaString nebelhornBaseUrl) pageDoc }

putSlug :: Page -> Page
putSlug page@Page{..} = case slug of
    Nothing -> page
    Just slug' -> page { pageDoc = addMeta "slug" (MetaString slug') pageDoc }
    where
        slug :: Maybe Text
        slug = pageTarget
            |> splitExtension
            |>> fst
            |>> toFilePath
            |>> toText
            |>> T.concatMap (\c -> if c `elem` ['/', '\\'] then "__" else T.singleton c)
            |>> T.map (\c -> if isAlphaNum c then c else '_')


-- | Reads category info from page document meta and puts it into the pageCategories
-- attribute.
putCategories :: Page -> Page
putCategories page@Page{..} = page
    {-
    pageCategories is pre-populated with the default categories.
    If the categories meta var is set, that overwrites the default ones.
    The extraCategories meta var is added on top of that.
    -}
    { pageCategories = fromMaybe pageCategories (getCategoriesFromMeta pageDoc)
        <> getExtraCategoriesFromMeta pageDoc
    } where

        getCategoriesFromMeta :: Pandoc -> Maybe [Text]
        getCategoriesFromMeta (Pandoc meta _) = case lookupMeta "categories" meta of
            Nothing -> Nothing  -- if nothing is set, we keep the default
            Just (MetaList vs) -> Just $ mapMaybe metaToPlainText vs
            _ -> Just []  -- if set to something invalid, we overwrite the default with an empty list

        getExtraCategoriesFromMeta :: Pandoc -> [Text]
        getExtraCategoriesFromMeta (Pandoc meta _) = case lookupMeta "extraCategories" meta of
            Just (MetaList vs) -> mapMaybe metaToPlainText vs
            _ -> []


-- | Sorts a list of pandocuments by their date meta values.
sortArticles :: Nebelhorn -> Nebelhorn
sortArticles nebelhorn@Nebelhorn{..} = nebelhorn { nebelhornPages = sortOn f nebelhornPages }
    where
        f = Down . (\(Pandoc meta _) -> lookupMeta "date" meta) . pageDoc

getCategory :: Text -> Nebelhorn -> [Page]
getCategory category Nebelhorn{..} = filter f nebelhornPages where
    f :: Page -> Bool
    f Page{..} = category `elem` pageCategories


-- | Puts prev and next tags into all articles.
-- Articles not for index are ignored, and also put in front of the list.
putNeighbors :: Nebelhorn -> Nebelhorn
putNeighbors = mapOnEachCategory putNeighbors' where

    putNeighbors' :: [Page] -> [Page]
    putNeighbors' = map insertNeighbors . getNeighbors

    -- TODO should be possible without maybe
    -- | Gets the previous and next neigbors for each article in the list.
    -- This only works if every pandocument has the link meta value set.
    getNeighbors :: [Page] -> [(Page, Maybe Text, Maybe Text)]
    getNeighbors pages = f <$> zip pages [0..] where
        f (p, n) = (p, getLink <$> pages `safeIndex` (n-1), getLink <$> pages `safeIndex` (n+1))

    getLink:: Page -> Text
    getLink Page{..} = toText $ toFilePath $ [absdir|/|] </> pageTarget

    insertNeighbors :: (Page, Maybe Text, Maybe Text) -> Page
    insertNeighbors (page, nextM, prevM)
        = applyDoc (addMeta "prev" (MetaString $ fromMaybe "" prevM))
        . applyDoc (addMeta "next" (MetaString $ fromMaybe "" nextM))
        $ page

    getCategories :: Nebelhorn -> [Text]
    getCategories Nebelhorn{..} = foldr union [] (pageCategories <$> nebelhornPages)

    putCategory :: Text -> [Page] -> Nebelhorn -> Nebelhorn
    putCategory category newPages nebelhorn@Nebelhorn{..} = nebelhorn { nebelhornPages = helper nebelhornPages newPages } where
        helper :: [Page] -> [Page] -> [Page]
        helper (page:pages) (newPage:newPages)
            | (head <$> nonEmpty page.pageCategories) == Just category = newPage : helper pages newPages
            | otherwise = page : helper pages (newPage:newPages)
        helper (page:pages) []
            | (head <$> nonEmpty page.pageCategories) == Just category = helper pages []
            | otherwise = page : helper pages []
        helper [] _ = []

    mapOnEachCategory :: ([Page] -> [Page]) -> Nebelhorn -> Nebelhorn
    mapOnEachCategory f nebelhorn = helper (getCategories nebelhorn) nebelhorn where
        helper :: [Text] -> Nebelhorn -> Nebelhorn
        helper [] nebelhorn = nebelhorn
        helper (c:cs) nebelhorn = helper cs $ putCategory c (f (getCategory c nebelhorn)) nebelhorn


indexFor :: Page -> Maybe Text
indexFor Page{..} = let Pandoc meta _ = pageDoc in lookupMetaPlain "indexFor" meta


putIsActive :: Page -> Page
putIsActive page@Page{..} = page { pageDoc = addMeta "isActive" isActiveMeta pageDoc } where

    isActiveMeta :: MetaValue
    isActiveMeta = MetaMap $ fromList
        [ ("index", isActiveMetaIndex)
        , ("category", isActiveMetaCategory)
        , ("page", isActiveMetaPage)
        ]

    -- isActive.index.foo is true if the current page is an index for the category foo
    isActiveMetaIndex :: MetaValue
    isActiveMetaIndex = case indexFor page of
        Nothing -> MetaMap Map.empty
        Just category -> MetaMap $ Map.singleton category $ MetaBool True

    -- isActive.category.foo is true for pages that have the category foo
    isActiveMetaCategory :: MetaValue
    isActiveMetaCategory = MetaMap $ fromList $ map helper pageCategories where

        helper :: Text -> (Text, MetaValue)
        helper category = (category, MetaMap $ Map.singleton category $ MetaBool True)

    -- isActive.page.foo is true if the current page has the slug foo
    isActiveMetaPage :: MetaValue
    isActiveMetaPage = let Pandoc meta _ = pageDoc in case lookupMetaPlain "slug" meta of
        Nothing -> MetaMap Map.empty
        Just slug -> MetaMap $ Map.singleton slug (MetaBool True)



putIndex :: Nebelhorn -> Nebelhorn
putIndex nebelhorn = mapPages putIndex' nebelhorn where

    putIndex' :: Page -> Page
    putIndex' page@Page{..} = case indexFor page of
        Nothing -> page
        Just category -> page { pageDoc = addMeta "indexPages"
            (MetaList $ getMetadata <$> getCategory category nebelhorn)
            pageDoc }

    -- | Generates the Pandoc MetaMap for a document for the index page.
    -- It just contains the metadata of the source.
    getMetadata :: Page -> MetaValue
    getMetadata Page{..} = let Pandoc (Meta metaMap) _ = pageDoc in MetaMap metaMap


applyFilters :: Nebelhorn -> IO Nebelhorn
applyFilters nebelhorn@Nebelhorn{..} = do
    pages' <- (runIOorExplode <$> apply) `mapConcurrently` nebelhornPages
    return nebelhorn { nebelhornPages = pages' }
    where

        apply :: (PandocMonad m, MonadIO m) => Page -> m Page
        apply page@Page{..} = do
            doc' <- Filter.applyFilters def nebelhornFilters ["html"] pageDoc
            return page { pageDoc = doc' }

putFeeds :: Nebelhorn -> Nebelhorn
putFeeds nebelhorn@Nebelhorn{..} = nebelhorn { nebelhornFeeds = putFeed <$> nebelhornFeeds } where

    putFeed :: Feed -> Feed
    putFeed feed@Feed{..} = feed { feedPages = filter (hasFeedCategory feedCategories) nebelhornPages }

    hasFeedCategory :: [Text] -> Page -> Bool
    hasFeedCategory feedCategories Page{..} = any (`elem` feedCategories) pageCategories


-- Builds a Nebelhorn page according to a config.
build :: Config -> Path Abs Dir -> IO ()
build config outputPath = liftIO (load config)
    >>= mapPages putSlug
    |>> mapPages linkIntoMeta
    |>> baseUrlIntoMeta
    |>> applyFilters
    >>= sortArticles
    |>> mapPages putCategories
    |>> mapPages putIsActive
    |>> putNeighbors
    |>> putIndex
    |>> putFeeds
    |>> save outputPath
