module Feed where

import Relude

import Nebelhorn
import Util

import Path
import Text.Pandoc.Definition
import qualified Text.Atom.Feed as Atom
import qualified Text.Atom.Feed.Export as Export
import Text.Pandoc (PandocMonad, runIOorExplode)
import Text.Pandoc.Writers (writeHtml5String)
import Text.Pandoc.Options (def)
import Data.Time
import Data.Time.Format.ISO8601 (iso8601Show)
import Data.Time.Calendar.OrdinalDate (fromOrdinalDate)
import Control.Concurrent.Async

-- See also https://validator.w3.org/feed/docs/atom.html

toEntry :: PandocMonad m => Text -> Page -> m Atom.Entry
toEntry baseUrl page@Page{..} = let Pandoc meta _ = pageDoc in do
    html <- renderPageContent page
    let entry = Atom.nullEntry
                (baseUrl <> fromMaybe "" (lookupMetaPlain "link" meta))
                (Atom.TextString $ fromMaybe "Untitled" (lookupMetaPlain "title" meta))
                timestamp
    return entry { Atom.entryContent = Just $ Atom.HTMLContent html }
    where
        timestamp :: Atom.Date
        timestamp = toText . iso8601Show $ getUpdated page

getUpdated :: Page -> ZonedTime
getUpdated Page{..} = case (pageDate, pageUpdated) of
        (_, Just updated) -> updated
        (Just date, _) -> date
        _ -> placeholderTime

placeholderTime :: ZonedTime
placeholderTime = ZonedTime (LocalTime (fromOrdinalDate 1980 1) midnight) utc

latestUpdate :: Feed -> ZonedTime
latestUpdate Feed{..} = fromMaybe placeholderTime $ viaNonEmpty head (getUpdated <$> sortOn (Down . zonedTimeToUTC . getUpdated) feedPages)

    -- maximum $ zonedTimeToUTC . getUpdated <$> feedPages

toFeed :: PandocMonad m => Text -> Feed -> m Atom.Feed
toFeed baseUrl feed@Feed{..} = do
    entries <- toEntry baseUrl `mapM` feedPages
    let entry = Atom.nullFeed
                (baseUrl <> (toText . toFilePath) ([absdir|/|] </> feedTarget))
                (Atom.TextString feedTitle)
                (toText . iso8601Show $ latestUpdate feed)
    return entry
        { Atom.feedEntries = entries
        , Atom.feedLinks = [Atom.nullLink baseUrl]
        }

{-renderFeed :: Atom.Feed -> Text
renderFeed feed = toStrict $ fromMaybe "" (Export.textFeed feed)-}

-- writes without template
renderPageContent :: (PandocMonad m) => Page -> m Text
renderPageContent Page{..} = writeHtml5String def pageDoc

renderFeed :: Atom.Feed -> Text
renderFeed feed = toStrict $ fromMaybe "" (Export.textFeed feed)

writeFeed :: (PandocMonad m, MonadIO m) => Path a Dir -> Text -> Feed -> m ()
writeFeed basePath baseUrl feed@Feed{..} = do
    let fileOutput = basePath </> feedTarget
    feed <- toFeed baseUrl feed
    let feedOut = renderFeed feed
    writeFileTextUtf8 (toFilePath fileOutput) feedOut

writeFeeds :: Path a Dir -> Nebelhorn -> IO ()
writeFeeds basePath Nebelhorn{..} =
    (runIOorExplode <$> writeFeed basePath nebelhornBaseUrl) `mapConcurrently_` nebelhornFeeds
