{-# OPTIONS_GHC -Wno-unrecognised-pragmas #-}
{-# HLINT ignore "Redundant <$>" #-}

module Nebelhorn (Nebelhorn(..), Page(..), Feed(..), CopyItem(..), load) where

import Relude
import Util
import Path
import Path.IO
import Config
import Text.Pandoc.Class
import Text.Pandoc.Definition
import Text.Pandoc.Options
import Text.Pandoc.Readers.Markdown
import Text.Pandoc.Templates
import Text.Pandoc.Shared (headerShift)
import qualified Text.Pandoc.Filter as Filter
import Data.Time
import Data.Time.TZInfo
import Data.Time.Format.ISO8601 (iso8601ParseM)
import Data.Text (strip, dropWhileEnd)
import Data.Time.TZTime (fromLocalTime, toZonedTime)
import Control.Concurrent.Async (mapConcurrently)

-- TODO: Refactor used used types to be more consistent
-- | The Nebelhorn type.
data Nebelhorn = Nebelhorn
    { nebelhornBaseUrl :: Text  -- TODO put into meta
    , nebelhornPages :: [Page]
    , nebelhornFeeds :: [Feed]
    , nebelhornCopyItems :: [CopyItem]  -- ^ The list of folders to just copy over.
    , nebelhornStylesheet :: FilePath  -- ^ The stylesheet content.
    , nebelhornFilters :: [Filter.Filter]
    }

data Page = Page
    { pageTarget :: Path Rel File
    , pageTemplate :: Template Text
    , pageDate :: Maybe ZonedTime
    , pageUpdated :: Maybe ZonedTime
    , pageCategories :: [Text]
    , pageDoc :: Pandoc
    }

data Feed = Feed
    { feedTitle :: Text
    , feedTarget :: Path Rel File
    , feedCategories :: [Text]
    , feedPages :: [Page]
    }

data CopyItem
    = CopyItemDir
        { copyItemDirSource :: Path Abs Dir
        , copyItemDirTarget :: Path Rel Dir
        }
    | CopyItemFile
        { copyItemFileSource :: Path Abs File
        , copyItemFileTarget :: Path Rel File
        }

-- | Reads in a file, always using UTF8 encoding.
-- Needed because Pandoc requires UTF8 input and readFileText
-- by default uses the system encoding which might be different,
-- leading to crashes.
readFileTextUtf8 :: MonadIO m => Path a File -> m Text
readFileTextUtf8  path = do
    pathAbs <- makeAbsolute path
    pathAbs |> toFilePath |> readFileBS |>> decodeUtf8

-- | Reads in a Nebelhorn value using a config.
load :: Config -> IO Nebelhorn
load Config{..} = Nebelhorn
    <$> return (dropWhileEnd (=='/'). strip $ configBaseUrl)
    <*> (concat <$> mapConcurrently readSource configSources)
    <*> return (getFeed <$> configFeeds)
    <*> mapM makeCopyItem configCopyItems
    <*> return "style/stylesheet.scss"  -- TODO make configurable
    <*> return (getFilter <$> configFilters)
    where

    makeCopyItem :: ConfigCopy -> IO CopyItem
    makeCopyItem ConfigCopy{..} = do
        srcDir <- resolveDir' configCopySource
        srcDirExists <- doesDirExist srcDir
        if srcDirExists then do
            outDir <- parseRelDir configCopyOutput
            return $ CopyItemDir srcDir outDir
        else do
            srcFile <- resolveFile' configCopySource
            outFile <- parseRelFile configCopyOutput
            return $ CopyItemFile srcFile outFile

    readSource :: ConfigSource -> IO [Page]
    readSource configSource@ConfigSource{..} = do
        srcDir <- resolveDir' configSourceSource
        srcDirExists <- doesDirExist srcDir
        if srcDirExists then do
            outDir <- parseRelDir configSourceOutput
            readSourceDir configSource srcDir outDir
        else do
            srcFile <- resolveFile' configSourceSource
            outFile <- parseRelFile configSourceOutput
            page <- readSourceFile configSource srcFile outFile
            return [page]

    readSourceDir :: ConfigSource -> Path Abs Dir -> Path Rel Dir -> IO [Page]
    readSourceDir ConfigSource{..} src out = do
        templateDefault <- readTemplate configSourceTemplate
        (_, filesAll) <- listDirRecurRel src
        files <- mapM fileHelper . filter isMarkdownFile $ filesAll
        mapConcurrently (readFile templateDefault) files
        where

            fileHelper :: Path Rel File -> IO (Path Abs File, Path Rel File)
            fileHelper file = do
                outFile <- replaceExtension ".html" file
                return (src </> file, out </> outFile)

            readFile :: Template Text -> (Path Abs File, Path Rel File) -> IO Page
            readFile templateDefault (srcFile, outFile) = do
                doc@(Pandoc meta _) <- readDoc srcFile
                template <- case lookupMetaPlain "template" meta of
                    -- there's no template set in the metadata, use the default for the source
                    Nothing -> return templateDefault
                    -- read the template file specified in the metadata
                    Just t -> parseSomeFile (toString t) >>= \case
                            Abs p -> readTemplate p
                            Rel p -> readTemplate p
                let (date, updated) = readTimes configTimezone doc
                return $ Page outFile template date updated configSourceDefaultCategories doc

    readSourceFile :: ConfigSource -> Path Abs File -> Path Rel File -> IO Page
    readSourceFile ConfigSource{..} src out = do
        doc@(Pandoc meta _) <- readDoc src
        let (date, updated) = readTimes configTimezone doc
        template <- case lookupMetaPlain "template" meta of
            -- there's no template set in the metadata, use the default for the source
            Nothing -> readTemplate configSourceTemplate
            -- read the template file specified in the metadata
            Just t -> parseSomeFile (toString t) >>= \case
                    Abs p -> readTemplate p
                    Rel p -> readTemplate p
        return $ Page out template date updated configSourceDefaultCategories doc

readTimes :: Text -> Pandoc -> (Maybe ZonedTime, Maybe ZonedTime)
readTimes timezone (Pandoc meta _) = (date, updated) where

    tzinfo :: TZInfo
    tzinfo = fromMaybe Data.Time.TZInfo.utc $ fromIdentifier timezone

    convertTime :: LocalTime -> ZonedTime
    convertTime = toZonedTime . fromLocalTime tzinfo

    dateStr, updatedStr :: Maybe Text
    dateStr = lookupMetaPlain "date" meta
    updatedStr = lookupMetaPlain "updated" meta

    date, updated :: Maybe ZonedTime
    date = convertTime <$> (parseDayOrTime =<< dateStr)
    updated = convertTime <$> (parseDayOrTime =<< updatedStr)




-- return (fromMaybe Data.Time.TZInfo.utc $ fromIdentifier configTimezone)

parseDayOrTime :: MonadFail m => Text -> m LocalTime
parseDayOrTime txt = case iso8601ParseM (toString txt) :: Maybe Day of
    Just day -> return $ LocalTime day midday
    Nothing -> iso8601ParseM (toString txt)

readDoc :: MonadIO m => Path Abs File -> m Pandoc
readDoc path = do
    putStrLn $ "reading " <> toFilePath path
    content <- readFileTextUtf8 path
    let opts = def {readerExtensions = pandocExtensions}
    liftIO (runIO (readMarkdown opts content)) >>= \case
        Left err -> do
            putStr "Skipping file: "
            print err
            return $ Pandoc (Meta $ fromList []) []
        Right doc -> return (headerShift 1 doc)

readTemplate :: (MonadIO m) => Path a File -> m (Template Text)
readTemplate path = do
    templateText <- readFileTextUtf8 path
    template <- liftIO $ compileTemplate (toFilePath path) templateText
    return $ either
        (\err -> error $ "template loading failed: " <> toText err)
        id
        template

-- | Predicate that checks whether a file is a markdown file using the file extension.
isMarkdownFile :: Path b File -> Bool
isMarkdownFile path = ext == ".md" || ext == ".markdown"
    where ext = fromMaybe "" $ fileExtension path

getFeed :: ConfigFeed -> Feed
getFeed ConfigFeed{..} = Feed configFeedTitle configFeedOutput configFeedCategories []

getFilter :: ConfigFilter -> Filter.Filter
getFilter ConfigFilter{..} = Filter.LuaFilter $ toFilePath configFilterFile
