-- | This module contains utility functions.
module Util where

import Relude
import Relude.Extra.Map (insert)
import Text.Pandoc.Definition
import System.IO.Unsafe
import Text.Pandoc (runPure)
import Text.Pandoc.Writers (writePlain)
import Text.Pandoc.Options (def)
import Data.Text (strip)

{- |
    A handy nested map function. Works just like the normal inline map '<$>', but nested.

    Example: @(+1) <$$> [[1, 2], [3, 4]]@ evaluates to @[[2, 3], [4, 5]]@.
-}
infixr 5 <$$>
(<$$>) :: (Functor f1, Functor f2) => (a -> b) -> f1 (f2 a) -> f1 (f2 b)
(<$$>) = fmap . fmap

{- |
Pipe operator that allows chaining function applications from left to right.

Example: @3 |> (+1) |> (*2) |> (+5)@
-}
infixl 8 |>
(|>) :: a -> (a -> b) -> b
x |> f = f x

infixl 8 |>>
(|>>) :: Functor f => f a -> (a -> b) -> f b
x |>> f = f <$> x

-- | The monadic equivalent of the nested map '<$$>'.
mapMT :: (Traversable t1, Traversable t2, Monad m) => (a -> m b) -> t1 (t2 a) -> m (t1 (t2 b))
mapMT = mapM . mapM

-- | Adds a key-metavalue pair to a pandocument. Already known keys are overwritten.
addMeta :: Text -> MetaValue -> Pandoc -> Pandoc
addMeta key value (Pandoc (Meta metaMap) bs)
    = Pandoc (Meta $ insert key value metaMap) bs

-- | Safely retrieve an item from a list, returning Nothing on failure.
-- Mostly equivalent to Data.Key.lookup, but returns Nothing on negative numbers which I rely on.
safeIndex :: [a] -> Int -> Maybe a
safeIndex [] _ = Nothing
safeIndex _ n | n < 0 = Nothing
safeIndex (x:_) 0 = Just x
safeIndex (_:xs) n = safeIndex xs (n-1)

{-# DEPRECATED oops "oops." #-}
oops :: Show a => Text -> a -> a
oops e a = unsafePerformIO (putText ("oops " <> e <> ": ") >> print a >> return a)

-- | Converts a MetaValue to plain text, if possible. Expects the meta value to be
-- a string or inlines.
metaToPlainText :: MetaValue -> Maybe Text
metaToPlainText = \case
    MetaString text -> Just text
    MetaInlines inlines -> strip <$> rightToMaybe (runPure (writePlain def (Pandoc nullMeta [Plain inlines])))
    _ -> Nothing

lookupMetaPlain :: Text -> Meta -> Maybe Text
lookupMetaPlain key meta = lookupMeta key meta >>= metaToPlainText

-- | Writes a file always using UTF8. Required due to Pandoc requiring UTF8 and
-- writeFileText using the system locale encoding.
writeFileTextUtf8 :: MonadIO m => FilePath -> Text -> m ()
writeFileTextUtf8 path = writeFileBS path . encodeUtf8
